import logging
from decimal import Decimal

from django.db import models
from django.db.models import QuerySet, Manager
from django.db.models.base import ModelBase
from django.db.models.manager import BaseManager
from django.db.models.query import ModelIterable
from django.forms import widgets as dj_widgets
from django.utils.translation import ugettext_lazy

from otree.common import expand_choice_tuples, get_app_label_from_import_path
from otree.currency import Currency, RealWorldCurrency
from . import idmap
from .vars import _PickleField, VarsMixin  # noqa

logger = logging.getLogger(__name__)


class OTreeModelBase(ModelBase):
    def __new__(mcs, name, bases, attrs):
        meta = attrs.get("Meta")
        module = attrs.get("__module__")
        is_concrete = not getattr(meta, "abstract", False)
        app_label = getattr(meta, "app_label", "")

        if is_concrete and module and not app_label:
            if meta is None:
                meta = type("Meta", (), {})
            app_label = get_app_label_from_import_path(module)
            meta.app_label = app_label
            meta.db_table = "{}_{}".format(app_label, name.lower())
            attrs["Meta"] = meta

        new_class = super().__new__(mcs, name, bases, attrs)

        for f in new_class._meta.fields:
            if hasattr(new_class, f.name + '_choices'):
                attr_name = 'get_%s_display' % f.name
                setattr(new_class, attr_name, make_get_display(f))

        new_class._setattr_fields = frozenset(f.name for f in new_class._meta.fields)
        new_class._setattr_attributes = frozenset(dir(new_class))

        return new_class


def make_get_display(field):
    def get_FIELD_display(self):
        choices = getattr(self, field.name + '_choices')()
        value = getattr(self, field.attname)
        return dict(expand_choice_tuples(choices))[value]

    return get_FIELD_display


class IdMapQuerySet(QuerySet):
    model = None  # type: OTreeModel

    def _idmap_clone(self):
        '''TODO: what is the purpose of this?'''
        clone = self._clone()
        clone.__dict__.update({'_fields': None, '_iterable_class': ModelIterable})
        return clone

    def get(self, *args, **kwargs):
        if not (idmap.is_active and self._iterable_class is ModelIterable):
            return super().get(*args, **kwargs)
        cache_kwargs = kwargs
        if len(args) == 1:
            # from django.db.models.query_utils import Q
            [q] = args
            cache_kwargs = dict(q.children)
        is_unsupported_lookup = bool(
            set(cache_kwargs.keys())
            - idmap.SUPPORTED_CACHE_LOOKUP_FIELDS[self.model.__name__]
        )

        instance = None
        if not is_unsupported_lookup:
            instance = self.model.get_cached_instance(**cache_kwargs)

        if instance is None:
            clone = self._idmap_clone()
            clone.query.clear_select_fields()
            clone.query.default_cols = True
            instance = super(IdMapQuerySet, clone).get(*args, **kwargs)
            if is_unsupported_lookup:
                # check if it already exists in cache
                cached_instance = self.model.get_cached_instance(id=instance.id)
                if cached_instance:
                    instance = cached_instance
            self.model.cache_if_necessary(instance)
        return instance


class IdMapManager(BaseManager.from_queryset(IdMapQuerySet), Manager):
    pass


class OTreeModel(models.Model, metaclass=OTreeModelBase):
    ####### IDMAP STUFF #######
    objects = IdMapManager()

    class Meta:
        abstract = True
        base_manager_name = 'objects'
        default_manager_name = 'objects'

    def __repr__(self):
        return '<{} pk={}>'.format(self.__class__.__name__, self.pk)

    _is_frozen = False
    NoneType = type(None)

    _setattr_datatypes = {
        # first value should be the "recommmended" datatype,
        # because that's what we recommend in the error message.
        # it seems the habit of setting boolean values to 0 or 1
        # is very common. that's even what oTree shows in the "live update"
        # view, and the data export.
        'BooleanField': (bool, int, NoneType),
        # forms seem to save Decimal to CurrencyField
        'CurrencyField': (Currency, NoneType, int, float, Decimal),
        'FloatField': (float, NoneType, int),
        'IntegerField': (int, NoneType),
        'StringField': (str, NoneType),
        'LongStringField': (str, NoneType),
    }
    _setattr_whitelist = {
        '_initial_prep_values',
        # used by Prefetch.
        '_ordered_players',
        '_is_frozen',
        # extras on 2018-11-24
        'id',
        '_changed_fields',
        'pk',
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # cache it for performance
        self._super_setattr = super().__setattr__
        # originally I had:
        # self._dir_attributes = set(dir(self))
        # but this tripled memory usage when creating a session
        # we still need some kind of 'save the change' logic because we don't want .update() queries
        # to be overwritten when the player model is saved.
        self._update_fields = set()
        self._is_frozen = True

    def __setattr__(self, field_name: str, value):
        if self._is_frozen:

            if field_name in self._setattr_fields:
                # hasattr() cannot be used inside
                # a Django model's setattr, as I discovered.
                field = self._meta.get_field(field_name)

                field_type_name = type(field).__name__
                if field_type_name in self._setattr_datatypes:
                    allowed_types = self._setattr_datatypes[field_type_name]
                    if (
                        isinstance(value, allowed_types)
                        # numpy uses its own datatypes, e.g. numpy._bool,
                        # which doesn't inherit from python bool.
                        or 'numpy' in str(type(value))
                        # 2018-07-18:
                        # have an exception for the bug in the 'quiz' sample game
                        # after a while, we can remove this
                        or field_name == 'question_id'
                    ):
                        pass
                    else:
                        msg = ('{} should be set to {}, not {}.').format(
                            field_type_name,
                            allowed_types[0].__name__,
                            type(value).__name__,
                        )
                        raise TypeError(msg)
            elif (
                field_name in self._setattr_attributes
                or field_name in self._setattr_whitelist
                or
                # idmap uses _group_cache, _subsession_cache,
                # _prefetched_objects_cache, etc
                field_name.endswith('_cache')
            ):
                # django sometimes reassigns to non-field attributes that
                # were set before the class was frozen, such as
                # .pk and ._changed_fields (from SaveTheChange)
                # or assigning to a property like Player.payoff
                pass
            else:
                msg = ('{} has no field "{}".').format(
                    self.__class__.__name__, field_name
                )
                raise AttributeError(msg)
            if field_name in self._setattr_fields and field_name not in ['id']:
                self._update_fields.add(field_name)
            self._super_setattr(field_name, value)
        else:
            # super() is a bit slower but only gets run during __init__
            super().__setattr__(field_name, value)

    def save(self, *args, **kwargs):
        if not self._state.adding:
            update_fields = self._update_fields
            if hasattr(self, '_vars_changed') and self._vars_changed():
                update_fields.add('vars')
            kwargs.setdefault('update_fields', list(self._update_fields))
        super().save(*args, **kwargs)
        # needed when we create a new group
        self.cache_if_necessary()

    def cache_if_necessary(self):
        if idmap.is_active:
            self.cache_instance(self)

    @classmethod
    def from_db(cls, db, field_names, values):
        '''i should figure out exactly when this is used. it apparently is required. maybe for FKs'''
        instance_id = dict(zip(field_names, values))['id']
        instance = cls.get_cached_instance(id=instance_id)
        if not instance:
            instance = super().from_db(db, field_names, values)
            cls.cache_if_necessary(instance)
        return instance

    def refresh_from_db(self, using=None, fields=None):
        self.flush_cached_instance(self)
        super().refresh_from_db(using, fields)
        self.cache_if_necessary()

    @classmethod
    def _get_cache_key(cls, id, **kwargs):
        return id

    @classmethod
    def get_cached_instance(cls, **kwargs):
        if 'pk' in kwargs:
            kwargs['id'] = kwargs.pop('pk')
        return cls._get_cached_instance(**kwargs)


def fix_choices_arg(kwargs):
    '''allows the programmer to define choices as a list of values rather
    than (value, display_value)

    '''
    choices = kwargs.get('choices')
    if not choices:
        return
    choices = expand_choice_tuples(choices)
    kwargs['choices'] = choices


class _OtreeModelFieldMixin:
    def __init__(
        self,
        *,
        initial=None,
        label=None,
        min=None,
        max=None,
        doc='',
        widget=None,
        **kwargs,
    ):

        self.widget = widget
        self.doc = doc
        self.min = min
        self.max = max

        fix_choices_arg(kwargs)

        kwargs.setdefault('help_text', '')
        kwargs.setdefault('null', True)

        # to be more consistent with {% formfield %}
        # this is more intuitive for newbies
        kwargs.setdefault('verbose_name', label)

        # "initial" is an alias for default. in the context of oTree, 'initial'
        # is a more intuitive name. (since the user never instantiates objects
        # themselves. also, "default" could be misleading -- people could think
        # it's the default choice in the form
        kwargs.setdefault('default', initial)

        # if default=None, Django will omit the blank choice from form widget
        # https://code.djangoproject.com/ticket/10792
        # that is contrary to the way oTree views blank/None values, so to
        # correct for this, we get rid of default=None args.
        # setting null=True already should make the field null
        if 'default' in kwargs and kwargs['default'] is None:
            kwargs.pop('default')

        super().__init__(**kwargs)

    def formfield(self, **kwargs):
        if self.widget:
            kwargs['widget'] = self.widget
        return super().formfield(**kwargs)


class _OtreeNumericFieldMixin(_OtreeModelFieldMixin):
    auto_submit_default = 0


class BaseCurrencyField(_OtreeNumericFieldMixin, models.DecimalField):

    MONEY_CLASS = None  # need to set in subclasses

    def __init__(self, **kwargs):
        # i think it's sufficient just to store a high number;
        # this needs to be higher than decimal_places
        decimal_places = self.MONEY_CLASS.get_num_decimal_places()
        # where does this come from?
        max_digits = 12
        super().__init__(max_digits=max_digits, decimal_places=decimal_places, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        # 2017-09-24: why do we need to do this?
        del kwargs["decimal_places"]
        del kwargs["max_digits"]
        return name, path, args, kwargs

    def to_python(self, value):
        value = models.DecimalField.to_python(self, value)
        if value is None:
            return value

        return self.MONEY_CLASS(value)

    def get_prep_value(self, value):
        if value is None:
            return None
        return Decimal(self.to_python(value))

    def from_db_value(self, value, expression, connection):
        return self.to_python(value)


class CurrencyField(BaseCurrencyField):
    MONEY_CLASS = Currency
    auto_submit_default = Currency(0)

    def formfield(self, **kwargs):
        import otree.forms

        defaults = {
            'form_class': otree.forms.CurrencyField,
            'choices_form_class': otree.forms.CurrencyChoiceField,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)


class RealWorldCurrencyField(BaseCurrencyField):
    MONEY_CLASS = RealWorldCurrency
    auto_submit_default = RealWorldCurrency(0)

    def formfield(self, **kwargs):
        import otree.forms

        defaults = {
            'form_class': otree.forms.RealWorldCurrencyField,
            'choices_form_class': otree.forms.CurrencyChoiceField,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)


class BooleanField(_OtreeModelFieldMixin, models.BooleanField):
    def __init__(self, **kwargs):
        # usually checkbox is not required, except for consent forms.
        widget = kwargs.get('widget')
        if isinstance(widget, dj_widgets.CheckboxInput):
            kwargs.setdefault('blank', True)

        # we need to set explicitly because otherwise the empty choice will show up as
        # "Unknown" in a select widget. This makes it set to '-----------'.
        kwargs.setdefault(
            'choices', [(True, ugettext_lazy('Yes')), (False, ugettext_lazy('No'))]
        )
        super().__init__(**kwargs)

    auto_submit_default = False


class StringField(_OtreeModelFieldMixin, models.CharField):
    '''
    Many people are already using initial=None, and i don't think it's
    causing any problems, even though Django recommends against that, but
    that's for forms on pages that get viewed multiple times
    '''

    def __init__(
        self,
        *,
        # varchar max length doesn't affect performance or even storage
        # size; it's just for validation. so, to be easy to use,
        # there is no reason for oTree to set a short default length
        # for CharFields. The main consideration is that MySQL cannot index
        # varchar longer than 255 chars, but that is not relevant here
        # because oTree only uses indexes for fields defined in otree-core,
        # which have explicit max_lengths anyway.
        max_length=10000,
        **kwargs,
    ):

        super().__init__(max_length=max_length, **kwargs)

    auto_submit_default = ''


# currently it's just a regular model, but we might customize it in the future
# (e.g. custom model manager)
ExtraModel = models.Model


class Link(models.ForeignKey):
    def __init__(self, to):
        # should we allow it to be null? That could be useful for network games.
        # or for games where we create resources that are initially unclaimed by users.
        # it's necessary for devserver auto migrations. if you add or rename a Link field,
        # the existing records must have null.
        # the biggest concern is that
        # people will forget to pass player or group.
        # but i guess i have no choice. anyway this is more advanced functionality.
        kwargs = dict(to=to, on_delete=models.CASCADE, null=True)
        # don't make reverse relation, then we don't have to worry about conflicting
        # related_name error.
        super().__init__(related_name="+", **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        new_kwargs = dict(to=kwargs['to'])
        return name, path, (), new_kwargs

    def formfield(self, *args, **kwargs):
        raise Exception('oTree Link cannot be used as formfield')


class DecimalField(_OtreeNumericFieldMixin, models.DecimalField):
    pass


Model = models.Model
ForeignKey = models.ForeignKey


class FloatField(_OtreeNumericFieldMixin, models.FloatField):
    pass


class IntegerField(_OtreeNumericFieldMixin, models.IntegerField):
    pass


class PositiveIntegerField(_OtreeNumericFieldMixin, models.PositiveIntegerField):
    pass


class LongStringField(_OtreeModelFieldMixin, models.TextField):
    auto_submit_default = ''


MSG_DEPRECATED_FIELD = """
{} does not exist in oTree. 
You should either replace it with one of oTree's field types, or import it from Django directly.
""".replace(
    '\n', ' '
)


def make_deprecated_field(FieldName):
    def DeprecatedField(*args, **kwargs):
        # putting the msg on a separate line gives better tracebacks
        raise Exception(MSG_DEPRECATED_FIELD.format(FieldName))

    return DeprecatedField


ManyToOneRel = make_deprecated_field("ManyToOneRel")
ManyToManyField = make_deprecated_field("ManyToManyField")
OneToOneField = make_deprecated_field("OneToOneField")
AutoField = make_deprecated_field("AutoField")
BigIntegerField = make_deprecated_field("BigIntegerField")
BinaryField = make_deprecated_field("BinaryField")
EmailField = make_deprecated_field("EmailField")
FileField = make_deprecated_field("FileField")
GenericIPAddressField = make_deprecated_field("GenericIPAddressField")
PositiveSmallIntegerField = make_deprecated_field("PositiveSmallIntegerField")
SlugField = make_deprecated_field("SlugField")
SmallIntegerField = make_deprecated_field("SmallIntegerField")
TimeField = make_deprecated_field("TimeField")
URLField = make_deprecated_field("URLField")
DateField = make_deprecated_field("DateField")
DateTimeField = make_deprecated_field("DateTimeField")


CharField = StringField
TextField = LongStringField
# keep ForeignKey around


CASCADE = models.CASCADE



import collections
import csv
import logging
import numbers
from collections import OrderedDict
from decimal import Decimal
from importlib import import_module
from django.utils.html import escape

import xlsxwriter
from django.db.models import BinaryField, ForeignKey
from django.db.models import Max
from django.utils.encoding import force_text

import otree
from otree.currency import Currency, RealWorldCurrency
from otree.common import get_models_module
from otree.common2 import TIME_SPENT_COLUMNS, write_page_completion_buffer
from otree.models.group import BaseGroup
from otree.models.participant import Participant
from otree.models.player import BasePlayer
from otree.models.session import Session
from otree.models.subsession import BaseSubsession
from otree.models_concrete import PageTimeBatch
from otree.session import SessionConfig

logger = logging.getLogger(__name__)


def inspect_field_names(Model):
    # filter out BinaryField, because it's not useful for CSV export or
    # live results. could be very big, and causes problems with utf-8 export

    # I tried .get_fields() instead of .fields, but that method returns
    # fields that cause problems, like saying group has an attribute 'player'
    field_names = []
    for f in Model._meta.fields:
        if not isinstance(f, BinaryField):
            if isinstance(f, ForeignKey):
                field_names.append('{}_id'.format(f.name))
            else:
                field_names.append(f.name)

    return field_names


def get_field_names_for_live_update(Model):
    return _get_table_fields(Model, for_export=False)


def get_field_names_for_csv(Model):
    return _get_table_fields(Model, for_export=True)


def _get_table_fields(Model, for_export=False):

    if Model is Session:
        # only data export
        return [
            'code',
            'label',
            'mturk_HITId',
            'mturk_HITGroupId',
            'comment',
            'is_demo',
        ]

    if Model is Participant:
        if for_export:
            return [
                'id_in_session',
                'code',
                'label',
                '_is_bot',
                '_index_in_pages',
                '_max_page_index',
                '_current_app_name',
                # this could be confusing because it will be in every row,
                # even rows for different rounds.
                #'_round_number',
                '_current_page_name',
                'time_started',
                'visited',
                'mturk_worker_id',
                'mturk_assignment_id',
                # last so that it will be next to payoff_plus_participation_fee
                'payoff',
            ]
        else:
            return [
                '_numeric_label',
                'code',
                'label',
                '_current_page',
                '_current_app_name',
                '_round_number',
                '_current_page_name',
                '_monitor_note',
                '_last_page_timestamp',
            ]

    if issubclass(Model, BasePlayer):
        subclass_fields = [
            f
            for f in inspect_field_names(Model)
            if f not in inspect_field_names(BasePlayer)
            and f not in ['id', 'group_id', 'subsession_id']
        ]

        return ['id_in_group', 'role'] + subclass_fields + ['payoff']

    if issubclass(Model, BaseGroup):
        subclass_fields = [
            f
            for f in inspect_field_names(Model)
            if f not in inspect_field_names(BaseGroup)
            and f not in ['id', 'subsession_id']
        ]

        return ['id_in_subsession'] + subclass_fields

    if issubclass(Model, BaseSubsession):
        subclass_fields = [
            f
            for f in inspect_field_names(Model)
            if f not in inspect_field_names(BaseGroup) and f != 'id'
        ]

        return ['round_number'] + subclass_fields


def sanitize_for_csv(value) -> str:
    if value is None:
        return ''
    if value is True:
        return 1
    if value is False:
        return 0
    if isinstance(value, (Currency, RealWorldCurrency)):
        # FIXME: django 1.11 upgrade: make sure querying currency values
        # doesn't slow it down too much (behavior of .values() changed)
        # Alexander has an idea, contact him if it's slow
        return Decimal(value)
    if isinstance(value, numbers.Number):
        return value
    value = force_text(value)
    return value.replace('\n', ' ').replace('\r', ' ')


def sanitize_for_live_update(value):
    # force_text is necessary e.g. for CountryField, which is otherwise
    # not Json serializable
    value = escape(sanitize_for_csv(value))
    MAX_LENGTH = 30
    if len(value) > MAX_LENGTH:
        return value[:MAX_LENGTH] + '…'
    return value


def get_payoff_plus_participation_fee(session, participant_values_dict):
    payoff = Currency(participant_values_dict['payoff'])
    return session._get_payoff_plus_participation_fee(payoff)


def get_rows_for_wide_csv(session_code):
    if session_code:
        sessions = [Session.objects.get(code=session_code)]
    else:
        sessions = Session.objects.order_by('id')
    participants = (
        Participant.objects.filter(session__in=sessions).order_by('id').values()
    )
    session_cache = {row.id: row for row in sessions}

    session_config_fields = set()
    for session in sessions:
        for field_name in SessionConfig(session.config).editable_fields():
            session_config_fields.add(field_name)
    session_config_fields = list(session_config_fields)

    if not participants:
        # 1 empty row
        return [[]]

    session_fields = get_field_names_for_csv(Session)
    participant_fields = get_field_names_for_csv(Participant)
    participant_fields.append('payoff_plus_participation_fee')
    header_row = ['participant.{}'.format(fname) for fname in participant_fields]
    header_row += ['session.{}'.format(fname) for fname in session_fields]
    header_row += ['session.config.{}'.format(fname) for fname in session_config_fields]
    rows = [header_row]
    for participant in participants:
        session = session_cache[participant['session_id']]
        participant[
            'payoff_plus_participation_fee'
        ] = get_payoff_plus_participation_fee(session, participant)
        row = [sanitize_for_csv(participant[fname]) for fname in participant_fields]

        row += [sanitize_for_csv(getattr(session, fname)) for fname in session_fields]
        row += [
            sanitize_for_csv(session.config.get(fname))
            for fname in session_config_fields
        ]
        rows.append(row)

    # heuristic to get the most relevant order of apps
    app_sequences = collections.Counter()
    for session in sessions:
        # we loaded the config earlier
        app_sequence = session.config['app_sequence']
        app_sequences[tuple(app_sequence)] += session.num_participants
    most_common_app_sequence = app_sequences.most_common(1)[0][0]

    # can't use settings.INSTALLED_OTREE_APPS, because maybe the app
    # was removed from SESSION_CONFIGS.
    app_names_with_data = set()
    for session in sessions:
        for app_name in session.config['app_sequence']:
            app_names_with_data.add(app_name)

    apps_not_in_popular_sequence = [
        app for app in app_names_with_data if app not in most_common_app_sequence
    ]

    order_of_apps = list(most_common_app_sequence) + apps_not_in_popular_sequence

    rounds_per_app = OrderedDict()
    for app_name in order_of_apps:
        try:
            models_module = get_models_module(app_name)
        except ModuleNotFoundError:
            # this should only happen with devserver because on production server,
            # you would need to resetdb after renaming an app.
            logger.warning(
                f'Cannot export data for app {app_name}, which existed when the session was run '
                f'but no longer exists.'
            )
            continue
        agg_dict = models_module.Subsession.objects.all().aggregate(Max('round_number'))
        highest_round_number = agg_dict['round_number__max']

        if highest_round_number is not None:
            rounds_per_app[app_name] = highest_round_number
    for app_name in rounds_per_app:
        for round_number in range(1, rounds_per_app[app_name] + 1):
            new_rows = get_rows_for_wide_csv_round(app_name, round_number, sessions)
            for i in range(len(rows)):
                rows[i].extend(new_rows[i])
    return rows


def get_rows_for_wide_csv_round(app_name, round_number, sessions):

    models_module = otree.common.get_models_module(app_name)
    Player = models_module.Player
    Group = models_module.Group
    Subsession = models_module.Subsession

    rows = []

    group_cache = {row['id']: row for row in Group.objects.values()}

    columns_for_models = {
        Model.__name__.lower(): get_field_names_for_csv(Model)
        for Model in [Player, Group, Subsession]
    }

    model_order = ['player', 'group', 'subsession']

    header_row = []
    for model_name in model_order:
        for colname in columns_for_models[model_name]:
            header_row.append(
                '{}.{}.{}.{}'.format(app_name, round_number, model_name, colname)
            )

    rows.append(header_row)
    empty_row = ['' for _ in range(len(header_row))]

    for session in sessions:
        subsession = Subsession.objects.filter(
            session_id=session.id, round_number=round_number
        ).values()
        if not subsession:
            subsession_rows = [empty_row for _ in range(session.num_participants)]
        else:
            subsession = subsession[0]
            subsession_id = subsession['id']
            players = (
                Player.objects.filter(subsession_id=subsession_id)
                .order_by('id')
                .values()
            )

            if len(players) != session.num_participants:
                msg = (
                    "Session {} has {} participants, but round {} of app '{}' "
                    "has {} players. The number of players in the subsession "
                    "should always match the number of players in the session. "
                    "Reset the database and examine your code."
                ).format(
                    session.code,
                    session.num_participants,
                    round_number,
                    app_name,
                    len(players),
                )
                raise AssertionError(msg)

            subsession_rows = []

            for player in players:
                # because these are properties
                player['payoff'] = player['_payoff']
                player['role'] = player['_role']
                row = []
                all_objects = {
                    'player': player,
                    'group': group_cache[player['group_id']],
                    'subsession': subsession,
                }

                for model_name in model_order:
                    for colname in columns_for_models[model_name]:
                        value = all_objects[model_name][colname]
                        row.append(sanitize_for_csv(value))
                subsession_rows.append(row)
        rows.extend(subsession_rows)
    return rows


def get_rows_for_csv(app_name):
    # need to use app_name and not app_label because the app might have been
    # removed from SESSION_CONFIGS
    models_module = otree.common.get_models_module(app_name)
    Player = models_module.Player
    Group = models_module.Group
    Subsession = models_module.Subsession

    columns_for_models = {
        Model.__name__.lower(): get_field_names_for_csv(Model)
        for Model in [Player, Group, Subsession, Participant, Session]
    }

    participant_ids = Player.objects.values_list('participant_id', flat=True)
    session_ids = Subsession.objects.values_list('session_id', flat=True)

    players = Player.objects.order_by('id').values()

    value_dicts = {
        'group': {row['id']: row for row in Group.objects.values()},
        'subsession': {row['id']: row for row in Subsession.objects.values()},
        'participant': {
            row['id']: row
            for row in Participant.objects.filter(id__in=participant_ids).values()
        },
        'session': {
            row['id']: row
            for row in Session.objects.filter(id__in=session_ids).values()
        },
    }

    model_order = ['participant', 'player', 'group', 'subsession', 'session']

    # header row
    rows = [
        [
            '{}.{}'.format(model_name, colname)
            for model_name in model_order
            for colname in columns_for_models[model_name]
        ]
    ]

    for player in players:
        # because these are properties
        player['payoff'] = player['_payoff']
        player['role'] = player['_role']
        row = []
        all_objects = {'player': player}
        for model_name in value_dicts:
            obj_id = player['{}_id'.format(model_name)]
            all_objects[model_name] = value_dicts[model_name][obj_id]

        for model_name in model_order:
            for colname in columns_for_models[model_name]:
                value = all_objects[model_name][colname]
                row.append(sanitize_for_csv(value))
        rows.append(row)

    return rows


def get_rows_for_monitor(participants) -> list:
    field_names = get_field_names_for_live_update(Participant)
    callable_fields = {'_numeric_label', '_current_page'}
    rows = []
    for participant in participants:
        row = {}
        for field_name in field_names:
            value = getattr(participant, field_name)
            if field_name in callable_fields:
                value = value()
            row[field_name] = value
        row['id_in_session'] = participant.id_in_session
        rows.append(row)
    return rows


def get_rows_for_live_update(subsession: BaseSubsession):

    models_module = import_module(subsession.__module__)
    Player = models_module.Player
    Group = models_module.Group
    Subsession = models_module.Subsession

    columns_for_models = {
        Model.__name__.lower(): get_field_names_for_live_update(Model)
        for Model in [Player, Group, Subsession]
    }

    # we had a strange result on one person's heroku instance
    # where Meta.ordering on the Player was being ingnored
    # when you use a filter. So we add one explicitly.
    players = (
        Player.objects.filter(subsession_id=subsession.pk)
        .select_related('group', 'subsession')
        .order_by('pk')
    )

    model_order = ['player', 'group', 'subsession']

    rows = []
    for player in players:
        row = []
        for model_name in model_order:
            if model_name == 'player':
                model_instance = player
            else:
                model_instance = getattr(player, model_name)

            for colname in columns_for_models[model_name]:

                attr = getattr(model_instance, colname, '')
                if callable(attr):
                    try:
                        attr = attr()
                    except Exception:
                        attr = ""
                row.append(sanitize_for_live_update(attr))
        rows.append(row)

    return columns_for_models, rows


def export_wide(fp, file_extension, session_code=None):
    rows = get_rows_for_wide_csv(session_code=session_code)
    _export_csv_or_xlsx(fp, rows, file_extension)


def export_app(app_name, fp, file_extension):
    rows = get_rows_for_csv(app_name)
    _export_csv_or_xlsx(fp, rows, file_extension)


def custom_export_app(app_name, fp, file_extension):
    models_module = get_models_module(app_name)
    qs = models_module.Player.objects.select_related(
        'participant', 'group', 'subsession', 'session'
    ).order_by('id')
    rows = models_module.custom_export(qs)
    # convert to strings so we don't get errors especially in xlsx write
    str_rows = []
    for row in rows:
        str_rows.append([str(ele) for ele in row])
    _export_csv_or_xlsx(fp, str_rows, file_extension)


def _export_csv_or_xlsx(fp, rows, file_extension):
    if file_extension == 'xlsx':
        _export_xlsx(fp, rows)
    else:
        _export_csv(fp, rows)


def _export_csv(fp, rows):
    writer = csv.writer(fp)
    writer.writerows(rows)


def _export_xlsx(fp, rows):
    '''
    CSV often does not open properly in Excel, e.g. unicode
    '''
    workbook = xlsxwriter.Workbook(fp, {'in_memory': True})
    worksheet = workbook.add_worksheet()

    for row_num, row in enumerate(rows):
        for col_num, cell_value in enumerate(row):
            worksheet.write(row_num, col_num, cell_value)
    workbook.close()


def export_page_times(fp):
    write_page_completion_buffer()
    batches = PageTimeBatch.objects.order_by('id').values_list('text', flat=True)
    fp.write(','.join(TIME_SPENT_COLUMNS) + '\n')
    for batch in batches:
        fp.write(batch)



from contextlib import contextmanager
from collections import defaultdict

participants_by_code = {}
participants_by_id = {}
sessions_by_code = {}
sessions_by_id = {}
subsessions = defaultdict(dict)
groups = defaultdict(dict)
players_by_id = defaultdict(dict)
players_by_participant_round = defaultdict(dict)

SUPPORTED_CACHE_LOOKUP_FIELDS = dict(
    Session=set(['code']),
    Participant=set(['code']),
    Subsession=set(),
    Group=set(),
    Player=set(['round_number', 'participant']),
    # See below! we add more stuff
)

for aset in SUPPORTED_CACHE_LOOKUP_FIELDS.values():
    aset.add('id')
    aset.add('pk')


def flush():
    for cache in [
        participants_by_id,
        participants_by_code,
        sessions_by_id,
        sessions_by_code,
        subsessions,
        groups,
        players_by_id,
        players_by_participant_round,
    ]:
        cache.clear()


def get_instances():
    for cache in [
        participants_by_id,
        sessions_by_id,
    ]:
        for instance in cache.values():
            yield instance
    for cache in [
        subsessions,
        groups,
        players_by_id,
    ]:
        for app_cache in cache.values():
            for instance in app_cache.values():
                yield instance


def save():
    for inst in get_instances():
        # object may have gotten deleted after being added to cache
        if inst.id:
            inst.save()


is_active = False


def activate():
    flush()
    global is_active
    is_active = True


def deactivate():
    flush()
    global is_active
    is_active = False


@contextmanager
def use_cache():
    activate()
    try:
        yield
        save()
    finally:
        deactivate()


class ParticipantIDMapMixin:
    @classmethod
    def _get_cached_instance(cls, id=None, code=None):
        assert id or code
        if id is not None:
            return participants_by_id.get(id)
        return participants_by_code.get(code)

    @classmethod
    def cache_instance(cls, instance):
        participants_by_id[instance.id] = instance
        participants_by_code[instance.code] = instance

    @classmethod
    def flush_cached_instance(cls, instance):
        participants_by_id.pop(instance.id, None)
        participants_by_code.pop(instance.code, None)


class SessionIDMapMixin:
    @classmethod
    def _get_cached_instance(cls, id=None, code=None):
        assert id or code
        if id is not None:
            return sessions_by_id.get(id)
        return sessions_by_code.get(code)

    @classmethod
    def cache_instance(cls, instance):
        sessions_by_id[instance.id] = instance
        sessions_by_code[instance.code] = instance

    @classmethod
    def flush_cached_instance(cls, instance):
        sessions_by_id.pop(instance.id, None)
        sessions_by_code.pop(instance.code, None)


class SubsessionIDMapMixin:
    @classmethod
    def cache(cls) -> dict:
        return subsessions[cls._meta.app_label]

    @classmethod
    def _get_cached_instance(cls, id):
        return cls.cache().get(id)

    @classmethod
    def cache_instance(cls, instance):
        cls.cache()[instance.id] = instance

    @classmethod
    def flush_cached_instance(cls, instance):
        cls.cache().pop(instance.id, None)


class GroupIDMapMixin:
    @classmethod
    def cache(cls) -> dict:
        return groups[cls._meta.app_label]

    @classmethod
    def _get_cached_instance(cls, id):
        return cls.cache().get(id)

    @classmethod
    def cache_instance(cls, instance):
        cls.cache()[instance.id] = instance

    @classmethod
    def flush_cached_instance(cls, instance):
        cls.cache().pop(instance.id, None)


class PlayerIDMapMixin:
    @classmethod
    def cache(cls) -> dict:
        return players_by_id[cls._meta.app_label]

    @classmethod
    def round_cache(cls) -> dict:
        return players_by_participant_round[cls._meta.app_label]

    @classmethod
    def _get_cached_instance(cls, id=None, participant=None, round_number=None):
        assert id or (participant and round_number)
        if id is not None:
            return cls.cache().get(id)
        return cls.round_cache().get((participant.id, round_number))

    @classmethod
    def cache_instance(cls, instance):
        cls.cache()[instance.id] = instance
        cls.round_cache()[instance.participant_id, instance.round_number] = instance

    @classmethod
    def flush_cached_instance(cls, instance):
        cls.cache().pop(instance.id, None)
        cls.round_cache().pop((instance.participant_id, instance.round_number), None)



from otree.db import models
from otree.common import (
    get_models_module,
    in_round,
    in_rounds,
    InvalidRoundError,
)
from otree import common
from otree.models.fieldchecks import ensure_field
import django.core.exceptions
from django.db import models as djmodels
from otree.constants import BaseConstants, get_role, get_roles
from otree.db.idmap import GroupIDMapMixin


class BaseGroup(models.OTreeModel, GroupIDMapMixin):
    class Meta:
        abstract = True
        index_together = ['session', 'id_in_subsession']
        ordering = ['pk']

    id_in_subsession = models.PositiveIntegerField(db_index=True)

    session = djmodels.ForeignKey(
        'otree.Session',
        related_name='%(app_label)s_%(class)s',
        on_delete=models.CASCADE,
    )

    round_number = models.PositiveIntegerField(db_index=True)

    @property
    def _Constants(self) -> BaseConstants:
        return get_models_module(self._meta.app_config.name).Constants

    def __unicode__(self):
        return str(self.pk)

    def get_players(self):
        return list(self.player_set.order_by('id_in_group'))

    def get_player_by_id(self, id_in_group):
        try:
            return self.player_set.get(id_in_group=id_in_group)
        except django.core.exceptions.ObjectDoesNotExist:
            msg = 'No player with id_in_group {}'.format(id_in_group)
            raise ValueError(msg) from None

    def get_player_by_role(self, role):
        if get_roles(self._Constants):
            try:
                return self.player_set.get(_role=role)
            except django.core.exceptions.ObjectDoesNotExist:
                pass
        else:
            for p in self.get_players():
                if p.role() == role:
                    return p
        msg = f'No player with role "{role}"'
        raise ValueError(msg)

    def set_players(self, players_list):
        Constants = self._Constants
        roles = get_roles(Constants)
        for i, player in enumerate(players_list, start=1):
            player.group = self
            player.id_in_group = i
            player._role = get_role(roles, i)
            player.save()

    def in_round(self, round_number):
        try:
            return in_round(
                type(self),
                round_number,
                session=self.session,
                id_in_subsession=self.id_in_subsession,
            )
        except InvalidRoundError as exc:
            msg = (
                str(exc)
                + '; '
                + (
                    'Hint: you should not use this '
                    'method if you are rearranging groups between rounds.'
                )
            )
            ExceptionClass = type(exc)
            raise ExceptionClass(msg) from None

    def in_rounds(self, first, last):
        try:
            return in_rounds(
                type(self),
                first,
                last,
                session=self.session,
                id_in_subsession=self.id_in_subsession,
            )
        except InvalidRoundError as exc:
            msg = (
                str(exc)
                + '; '
                + (
                    'Hint: you should not use this '
                    'method if you are rearranging groups between rounds.'
                )
            )
            ExceptionClass = type(exc)
            raise ExceptionClass(msg) from None

    def in_previous_rounds(self):
        return self.in_rounds(1, self.round_number - 1)

    def in_all_rounds(self):
        return self.in_previous_rounds() + [self]

    @classmethod
    def _ensure_required_fields(cls):
        """
        Every ``Group`` model requires a foreign key to the ``Subsession``
        model of the same app.
        """
        subsession_model = '{app_label}.Subsession'.format(
            app_label=cls._meta.app_label
        )
        subsession_field = djmodels.ForeignKey(
            subsession_model, on_delete=models.CASCADE
        )
        ensure_field(cls, 'subsession', subsession_field)



import re
import sys
from pathlib import Path

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    something_changed = False

    def add_arguments(self, parser):
        ahelp = 'Tells the command to NOT prompt the user for ' 'input of any kind.'
        parser.add_argument(
            '--noinput',
            action='store_false',
            dest='interactive',
            default=True,
            help=ahelp,
        )

    def handle(self, *args, **options):

        self.stdout = sys.stdout
        self.dry_run = options["interactive"]

        if self.dry_run:
            self.scan()
            if self.something_changed:
                prompt = (
                    'You should back up your code before you proceed. '
                    'Proceed with syntax upgrade? (y or n): '
                )
                confirmed = input(prompt).lower() == 'y'
                if not confirmed:
                    sys.exit()
            else:
                self.stdout.write('Code is up to date; no changes to be made.')
                sys.exit()

        self.dry_run = False
        self.scan()

    def scan(self):

        html_fns = Path('.').glob('**/*.html')
        for fn in html_fns:
            self.apply_rule_to_file(
                fn, r"% formfield (player|group)\.(\w+)", r"% formfield '\2'"
            )

    def apply_rule_to_file(self, fp: Path, before_pattern: str, after_pattern: str):
        something_changed = False
        new_lines = []
        for i, line in enumerate(fp.open(encoding='utf-8')):
            new_line = re.sub(before_pattern, after_pattern, line)
            if new_line != line:
                self.something_changed = True
                something_changed = True
                if self.dry_run:
                    self.print_change('{}, line {}\n'.format(fp, i + 1))
                    self.print_change('BEFORE: {}'.format(line))
                    self.print_change('AFTER:  {}'.format(new_line))
            new_lines.append(new_line)
        if something_changed and not self.dry_run:
            fp.write_text(''.join(new_lines))

    def print_change(self, msg):
        self.something_changed = True
        self.stdout.write(msg)



from django.apps import AppConfig

#
# # We import this here to ensure the reactor is installed very early on
# # in case other packages accidentally import twisted.internet.reactor
# # (e.g. raven does this).
# import daphne.server
#
# assert daphne.server  # pyflakes doesn't support ignores
#


class ChannelsConfig(AppConfig):

    name = "channels"
    verbose_name = "Channels"

    def ready(self):
        pass
        # oTree: removed because i think not necessary for us, see the note inside that function.
        # Do django monkeypatches
        # from .hacks import monkeypatch_django
        # monkeypatch_django()
